/**
 * @file version.h
 * @brief Semantic version macros and runtime version queries for libhisto.
 */

#ifndef LIBHISTO_VERSION_H
#define LIBHISTO_VERSION_H

/**
 * @file version.h
 * @brief Version definitions for libhisto.
 */

/** @brief Major version number */
#define HISTO_VERSION_MAJOR 0

/** @brief Minor version number */
#define HISTO_VERSION_MINOR 3

/** @brief Patch version number */
#define HISTO_VERSION_PATCH 0

/** @brief String representation of the version */
#define HISTO_VERSION_STRING "0.3.0"

#endif /* LIBHISTO_VERSION_H */
